/**************************************************************************
 *	  	 Definitions           					  *
 **************************************************************************/
#ifndef DEFS_H
#define DEFS_H

#include <stdio.h>
#include <fstream>

/* Definition of constants */
#define MAXPOPSIZE      200   // max population size
#define	MAXCHROMLENGTH	100   // maximum chromosome length

/* Definitions */
typedef int allele; 		// type of allele 
typedef struct struture_individual {
  allele chromosome[MAXCHROMLENGTH]; // chromosome
  double fitness;		     // function evaluation value	
  int parent1, parent2;		     // indices for parents 	
} individual;

typedef struct struture_population {
  individual ind[MAXPOPSIZE];
  double sum_fitness;
  double mean_fitness;
  double max_fitness;
  int best_individual;
  int best_individual_2;	
} population;

/* Definition of global variables */
extern population popcurr, popnew;		
extern int gaMode; // GA mode: 1. standard (SGA), 
                   //  2. standard random immigrants (RIGA1) 
                   //  3. random immigrants replace the worst individuals (RIGA2)
                   //  4. elitism-based GA (EIGA)
                   //  5. random and elitism-based immigrants GA (ERIGA)
                   //  6. hybrid immigrants GA (HIGA)

extern int totalRuns;	// number of runs of a GA
extern int popsize, lchrom, maxgen;
extern int elitism;		  // 1: elitism ; 0: no elitism
extern int flag_no_crossover;	  // disable crossover
extern int flag_no_mutation;	  // disable mutation
extern int num_immigrants;	  // number of individuals replaced each generation
extern int num_rand_immigrants;   // number of random immigrants
extern int num_elite_immigrants;  // number of elitism-based immigrants
extern int num_dual_immigrants;   // number of dualism-based immigrants
extern int delta_immigrants;      // change of immigrants number
extern int min_immigrants;        // minimum number of immigrants for each scheme
extern double ratio_immigrants;   // ratio of immigrants to pop size
extern double pImmigMut;          // prob. of mutating the elite from last 
                                  // generation or its dual

extern int flag_elitism;
extern individual individual_best;
extern individual individual_best_2;
extern individual elite_prev_pop;    // elite from previous population
extern long int currGenNo;           // current generation number
extern double pCrossover, pMutation; // probabilities of crossover and mutation
extern double *log_mean_fitness;
extern double *log_best_fitness; 
extern double *log_mean_diversity; 
extern bool bRecordDiversity;
extern double tournament_param; // probability that the fitter individual wins 
                                // in the tournament selection

extern int functNo;  // Function: 1-DUF(25,4,1), 2-DUF(25,4,3), 3-DUF(25,4,4), 
                     //  4-DUF(25,4,deceptive), 25 copies of 4-bit deceptive problems
                     //  5-Knapsack Problem

extern int changeMode; // how environment changes, cyclical or not
                       //   1: cyclic, 2: cyclic with noise, 3: random
extern double changeDegree; // ratio of loci changed in dynamic environment
extern int changeSpeed;     // period of environment changing in generations
extern int totalChanges;    // total number of environment changes in a run
extern bool changeDetected; // whether the enviroment change has been detected
extern int *xorMask;        // mask for XORing environment
extern bool bRandRatio;     // whether changeDegree is random (false by default)
extern int **baseXORMasks;  // base xor masks for cyclic environment
extern int totalBaseStates;
extern double pBaseMut;     // prob. of mutating each element of a base XOR mask 

// Parameters for Knapsack problem
extern int *weight, *profit;
extern int CAPACITY, TOTALWEIGHT;
extern double capFactor;

/* Declaration of functions */
void ga(int num_exec);
void initialization(void);
void statistics(population *pop);
void calcPopDiversity(population *pop);
double calcFitness(int *ind);
void generation(void);
void crossover(allele *parent1, allele *parent2, 
               allele *child1, allele *child2);
int mutation(allele value, double pMut);
int selection(population *pop, int j);
int selection_roulette(population *pop, double sum_fitness);
int selection_tournament(population *pop);

int randomInteger(int low, int up);
double randomNumber(double low, double up);
int randomBit();
int sign(double value);
void swap(int &x, int &y);
void swap(double &x, double &y);
int max(int x, int y);
void randomSequence(int* seqArr, int seqSize, int numSize);
void initializeKnapsack(void);

void display(population *pop);
void changeEnvironment(void);
void constructBaseXORMasks(void);
void saveFitnessResult(int num_exec);
void saveDiversityResult(int num_exec);
void substitution(population *pop);

#endif
