/***************************************************************************
 *				Immigrants Schemes			   *
 ***************************************************************************/
#include "defs.h"
#include <stdio.h>
#include <stdlib.h>

#include <iostream>
#include <string>
#include <fstream>

using namespace std;

/****************************************************************************
 *  			Standard Random Immigrants			    *
 ****************************************************************************/
void imig_random(population *pop) {
  int i, idx, gene;

  for (i=0; i<num_immigrants; i++) {
    if (elitism == 1) {
      idx = pop->best_individual;
      while (idx == pop->best_individual || 
             idx == pop->best_individual_2)
	idx = randomInteger(0, popsize);   
    }
    else 
      idx = randomInteger(0, popsize);
 
    pop->sum_fitness -= pop->ind[idx].fitness;	
    for (gene=0; gene<lchrom; gene++)
      pop->ind[idx].chromosome[gene] = randomBit();

    pop->ind[idx].fitness = calcFitness( pop->ind[idx].chromosome );
    pop->sum_fitness += pop->ind[idx].fitness;
    pop->mean_fitness = pop->sum_fitness/popsize;

    if (pop->ind[idx].fitness > pop->max_fitness) {
      pop->best_individual = idx;
      pop->max_fitness = pop->ind[idx].fitness;
    }
    else {
      if (pop->ind[idx].fitness > pop->ind[pop->best_individual_2].fitness)
       pop->best_individual_2 = idx;  
    }
  }
}

/****************************************************************************
 *	Random Immigrants with replacement of the worst individuals         *
 ****************************************************************************/
void imig_worst(population *pop) {
  int i, j, gene, *vector_order, index_worst;

  vector_order = new int[popsize];

  // Selection sorting of the num_immigrants worst individual
  for (i=0; i<popsize; i++) vector_order[i] = i;
  
  for (i=0; i<num_immigrants; i++){
    index_worst = i;
    for (j=i+1; j<popsize; j++){
      if (pop->ind[vector_order[j]].fitness <= 
          pop->ind[vector_order[index_worst]].fitness) {
        index_worst = j;
      }
    }
    swap(vector_order[i], vector_order[index_worst]);
  }

  for (i=0; i<num_immigrants; i++) {	
    pop->sum_fitness -= pop->ind[vector_order[i]].fitness;	

    for (gene=0; gene<lchrom; gene++) 
      pop->ind[vector_order[i]].chromosome[gene] = randomBit();
			
    pop->ind[vector_order[i]].fitness = 
        calcFitness(pop->ind[vector_order[i]].chromosome);
    pop->sum_fitness += pop->ind[vector_order[i]].fitness;

    if ( pop->ind[vector_order[i]].fitness > pop->max_fitness ){
	pop->best_individual = vector_order[i];
	pop->max_fitness = pop->ind[vector_order[i]].fitness ;
    }
    else {
      if (pop->ind[vector_order[i]].fitness > 
          pop->ind[pop->best_individual_2].fitness) {
        pop->best_individual_2 = vector_order[i];
      }
    }
  }

  pop->mean_fitness = pop->sum_fitness/popsize;

  delete [] vector_order;
}

/****************************************************************************
 *	Elitism based immigrants                                            *
 ****************************************************************************/
void imig_elitism(population *pop) {
  int i, j, gene, *vector_order, index_worst;

  vector_order = new int[popsize];

  // Selection sorting of the num_immigrants worst individual
  for (i=0; i<popsize; i++) vector_order[i] = i;

  for (i=0; i<num_immigrants; i++) {
    index_worst = i;
    for (j=i+1; j<popsize; j++){
      if (pop->ind[vector_order[j]].fitness <= 
          pop->ind[vector_order[index_worst]].fitness) {
        index_worst = j;
      }
    }
    swap(vector_order[i], vector_order[index_worst]);
  }

  for (i=0; i<num_immigrants; i++) {	
    pop->sum_fitness -= pop->ind[vector_order[i]].fitness;	

    for (gene=0; gene<lchrom; gene++) 
      pop->ind[vector_order[i]].chromosome[gene] 
          = mutation(elite_prev_pop.chromosome[gene], pImmigMut);

    pop->ind[vector_order[i]].fitness = 
        calcFitness(pop->ind[vector_order[i]].chromosome);
    pop->sum_fitness += pop->ind[vector_order[i]].fitness;

    if ( pop->ind[vector_order[i]].fitness > pop->max_fitness ){
	pop->best_individual = vector_order[i];
	pop->max_fitness = pop->ind[vector_order[i]].fitness ;
    }
    else {
      if (pop->ind[vector_order[i]].fitness > 
          pop->ind[pop->best_individual_2].fitness) {
        pop->best_individual_2 = vector_order[i];
      }
    }
  }

  pop->mean_fitness = pop->sum_fitness/popsize;

  delete [] vector_order;
}

/****************************************************************************
 *	Random and elitism based immigrants                                 *
 ****************************************************************************/
void imig_elitism_random(population *pop) {
  int i, j, gene, *vector_order, index_worst;
  double best_EI_fitness = 0.0;
  double best_RI_fitness = 0.0;

  vector_order = new int[popsize];

  // Selection sorting of the num_immigrants worst individual
  for (i=0; i<popsize; i++) vector_order[i] = i;
  
  for (i=0; i<num_immigrants; i++) {
    index_worst = i;
    for (j=i+1; j<popsize; j++) {
      if (pop->ind[vector_order[j]].fitness <= 
          pop->ind[vector_order[index_worst]].fitness) {
        index_worst = j;
      }
    }
    swap(vector_order[i], vector_order[index_worst]);
  }

  // random immigrants
  for (i=0; i<num_rand_immigrants; i++) {	
    pop->sum_fitness -= pop->ind[vector_order[i]].fitness;	

    for (gene=0; gene<lchrom; gene++) 
      pop->ind[vector_order[i]].chromosome[gene] = randomBit();
			
    pop->ind[vector_order[i]].fitness = 
        calcFitness(pop->ind[vector_order[i]].chromosome);
    pop->sum_fitness += pop->ind[vector_order[i]].fitness;

    if (pop->ind[vector_order[i]].fitness > best_RI_fitness)
      best_RI_fitness = pop->ind[vector_order[i]].fitness;

    if (pop->ind[vector_order[i]].fitness > pop->max_fitness) {
	pop->best_individual = vector_order[i];
	pop->max_fitness = pop->ind[vector_order[i]].fitness;
    }
    else {
      if (pop->ind[vector_order[i]].fitness > 
          pop->ind[pop->best_individual_2].fitness) {
        pop->best_individual_2 = vector_order[i];
      }
    }
  }

  // elitsim-based immigrants
  for (i=num_rand_immigrants; i<num_immigrants; i++) {	
    pop->sum_fitness -= pop->ind[vector_order[i]].fitness;	

    for (gene=0; gene<lchrom; gene++) 
      pop->ind[vector_order[i]].chromosome[gene] = 
          mutation(elite_prev_pop.chromosome[gene], pImmigMut);

    pop->ind[vector_order[i]].fitness = 
        calcFitness(pop->ind[vector_order[i]].chromosome);
    pop->sum_fitness += pop->ind[vector_order[i]].fitness;

    if (pop->ind[vector_order[i]].fitness > best_EI_fitness)
      best_EI_fitness = pop->ind[vector_order[i]].fitness;

    if ( pop->ind[vector_order[i]].fitness > pop->max_fitness ){
	pop->best_individual = vector_order[i];
	pop->max_fitness = pop->ind[vector_order[i]].fitness ;
    }
    else {
      if (pop->ind[vector_order[i]].fitness > 
          pop->ind[pop->best_individual_2].fitness) {
        pop->best_individual_2 = vector_order[i];
      }
    }
  }

  // adjust the number of random and elitism-based immigrants
  if (best_EI_fitness > best_RI_fitness) {
    num_rand_immigrants = max(num_rand_immigrants - delta_immigrants, min_immigrants);
    num_elite_immigrants = num_immigrants - num_rand_immigrants;
  } else if (best_EI_fitness < best_RI_fitness) {
    num_elite_immigrants = max(num_elite_immigrants - delta_immigrants, min_immigrants);
    num_rand_immigrants = num_immigrants - num_elite_immigrants;
  }

  pop->mean_fitness = pop->sum_fitness/popsize;

  delete [] vector_order;
}

/****************************************************************************
 *	Dualism and elitism based immigrants                                *
 ****************************************************************************/
void imig_dualism(population *pop) {
  int i, j, gene, *vector_order, index_worst;
  double best_EI_fitness = 0.0;
  double best_DI_fitness = 0.0;

  vector_order = new int[popsize];

  // Selection sorting of the num_immigrants worst individual
  for (i=0; i<popsize; i++) vector_order[i] = i;
  
  for (i=0; i<num_immigrants; i++) {
    index_worst = i;
    for (j=i+1; j<popsize; j++) {
      if (pop->ind[vector_order[j]].fitness <= 
          pop->ind[vector_order[index_worst]].fitness) {
        index_worst = j;
      }
    }
    swap(vector_order[i], vector_order[index_worst]);
  }

  // elitsim-based immigrants
  for (i=0; i<num_elite_immigrants; i++) {	
    pop->sum_fitness -= pop->ind[vector_order[i]].fitness;	

    for (gene=0; gene<lchrom; gene++) 
      pop->ind[vector_order[i]].chromosome[gene] = 
          mutation(elite_prev_pop.chromosome[gene], pImmigMut);

    pop->ind[vector_order[i]].fitness = 
        calcFitness(pop->ind[vector_order[i]].chromosome);
    pop->sum_fitness += pop->ind[vector_order[i]].fitness;
   
    if (pop->ind[vector_order[i]].fitness > best_EI_fitness)
      best_EI_fitness = pop->ind[vector_order[i]].fitness;
    
    if (pop->ind[vector_order[i]].fitness > pop->max_fitness){
	pop->best_individual = vector_order[i];
	pop->max_fitness = pop->ind[vector_order[i]].fitness;
    }
    else {
      if (pop->ind[vector_order[i]].fitness > 
          pop->ind[pop->best_individual_2].fitness) {
        pop->best_individual_2 = vector_order[i];
      }
    }
  }

  // dualism-based immigrants
  for (i=num_elite_immigrants; i<num_immigrants; i++) {	
    pop->sum_fitness -= pop->ind[vector_order[i]].fitness;	

    for (gene=0; gene<lchrom; gene++) 
      pop->ind[vector_order[i]].chromosome[gene] = 
          mutation(elite_prev_pop.chromosome[gene], 1.0-pImmigMut);

    pop->ind[vector_order[i]].fitness = 
        calcFitness(pop->ind[vector_order[i]].chromosome);
    pop->sum_fitness += pop->ind[vector_order[i]].fitness;

    if (pop->ind[vector_order[i]].fitness > best_DI_fitness)
      best_DI_fitness = pop->ind[vector_order[i]].fitness;

    if ( pop->ind[vector_order[i]].fitness > pop->max_fitness ){
	pop->best_individual = vector_order[i];
	pop->max_fitness = pop->ind[vector_order[i]].fitness ;
    }
    else {
      if (pop->ind[vector_order[i]].fitness > 
          pop->ind[pop->best_individual_2].fitness) {
        pop->best_individual_2 = vector_order[i];
      }
    }
  }

  // adjust the number of elitism and dualism based immigrants
  if (best_EI_fitness > best_DI_fitness) {
    num_dual_immigrants = max(num_dual_immigrants - delta_immigrants, min_immigrants);
    num_elite_immigrants = num_immigrants - num_dual_immigrants;
  } else if (best_EI_fitness < best_DI_fitness) {
    num_elite_immigrants = max(num_elite_immigrants - delta_immigrants, min_immigrants);
    num_dual_immigrants = num_immigrants - num_elite_immigrants;
  }

  pop->mean_fitness = pop->sum_fitness/popsize;

  delete [] vector_order;
}

/****************************************************************************
 *	Hydrid immigrants: random, elitism, dualism based immigrants        *
 ****************************************************************************/
void imig_hybrid(population *pop) {
  int i, j, gene, *vector_order, index_worst;
  double best_RI_fitness  = 0.0;
  double best_EI_fitness = 0.0;
  double best_DI_fitness = 0.0;

  vector_order = new int[popsize];

  // Selection sorting of the num_immigrants worst individual
  for (i=0; i<popsize; i++) vector_order[i] = i;
  
  for (i=0; i<num_immigrants; i++){
    index_worst = i;
    for (j=i+1; j<popsize; j++){
      if (pop->ind[vector_order[j]].fitness <= 
          pop->ind[vector_order[index_worst]].fitness) {
        index_worst = j;
      }
    }
    swap(vector_order[i], vector_order[index_worst]);
  }

  // random immigrants
  for (i=0; i<num_rand_immigrants; i++) {	
    pop->sum_fitness -= pop->ind[vector_order[i]].fitness;	

    for (gene=0; gene<lchrom; gene++) 
      pop->ind[vector_order[i]].chromosome[gene] = randomBit();
			
    pop->ind[vector_order[i]].fitness = 
        calcFitness(pop->ind[vector_order[i]].chromosome);
    pop->sum_fitness += pop->ind[vector_order[i]].fitness;

    if (pop->ind[vector_order[i]].fitness > best_RI_fitness)
      best_RI_fitness = pop->ind[vector_order[i]].fitness;

    if (pop->ind[vector_order[i]].fitness > pop->max_fitness) {
	pop->best_individual = vector_order[i];
	pop->max_fitness = pop->ind[vector_order[i]].fitness;
    }
    else {
      if (pop->ind[vector_order[i]].fitness > 
          pop->ind[pop->best_individual_2].fitness) {
        pop->best_individual_2 = vector_order[i];
      }
    }
  }

  // elitsim-based immigrants
  for (i=num_rand_immigrants; 
       i<num_rand_immigrants + num_elite_immigrants; i++) {	
    pop->sum_fitness -= pop->ind[vector_order[i]].fitness;	

    for (gene=0; gene<lchrom; gene++) 
      pop->ind[vector_order[i]].chromosome[gene] = 
          mutation(elite_prev_pop.chromosome[gene], pImmigMut);

    pop->ind[vector_order[i]].fitness = 
        calcFitness(pop->ind[vector_order[i]].chromosome);
    pop->sum_fitness += pop->ind[vector_order[i]].fitness;

    if (pop->ind[vector_order[i]].fitness > best_EI_fitness)
      best_EI_fitness = pop->ind[vector_order[i]].fitness;

    if ( pop->ind[vector_order[i]].fitness > pop->max_fitness ){
	pop->best_individual = vector_order[i];
	pop->max_fitness = pop->ind[vector_order[i]].fitness ;
    }
    else {
      if (pop->ind[vector_order[i]].fitness > 
          pop->ind[pop->best_individual_2].fitness) {
        pop->best_individual_2 = vector_order[i];
      }
    }
  }

  // dualism-based immigrants
  for (i=num_rand_immigrants + num_elite_immigrants; i<num_immigrants; i++) {
    pop->sum_fitness -= pop->ind[vector_order[i]].fitness;	

    for (gene=0; gene<lchrom; gene++) 
      pop->ind[vector_order[i]].chromosome[gene] = 
          mutation(elite_prev_pop.chromosome[gene], 1.0-pImmigMut);

    pop->ind[vector_order[i]].fitness = 
        calcFitness(pop->ind[vector_order[i]].chromosome);
    pop->sum_fitness += pop->ind[vector_order[i]].fitness;

    if (pop->ind[vector_order[i]].fitness > best_DI_fitness)
      best_DI_fitness = pop->ind[vector_order[i]].fitness;

    if ( pop->ind[vector_order[i]].fitness > pop->max_fitness ){
	pop->best_individual = vector_order[i];
	pop->max_fitness = pop->ind[vector_order[i]].fitness ;
    }
    else {
      if (pop->ind[vector_order[i]].fitness > 
          pop->ind[pop->best_individual_2].fitness) {
        pop->best_individual_2 = vector_order[i];
      }
    }
  }

  // adjust the number of random, elitism, and dualism based immigrants
  if (best_RI_fitness > best_EI_fitness) {
    if (best_RI_fitness > best_DI_fitness) {
      num_elite_immigrants = max(num_elite_immigrants - delta_immigrants, min_immigrants);
      num_dual_immigrants = max(num_dual_immigrants - delta_immigrants, min_immigrants);
      num_rand_immigrants = num_immigrants - num_elite_immigrants - num_dual_immigrants;
    } else if (best_RI_fitness < best_DI_fitness) {
      num_rand_immigrants = max(num_rand_immigrants - delta_immigrants, min_immigrants);
      num_elite_immigrants = max(num_elite_immigrants - delta_immigrants, min_immigrants);
      num_dual_immigrants = num_immigrants - num_rand_immigrants - num_elite_immigrants;
    }
  } else if (best_RI_fitness < best_EI_fitness) {
    if (best_EI_fitness > best_DI_fitness) {
      num_rand_immigrants = max(num_rand_immigrants - delta_immigrants, min_immigrants);
      num_dual_immigrants = max(num_dual_immigrants - delta_immigrants, min_immigrants);
      num_elite_immigrants = num_immigrants - num_rand_immigrants - num_dual_immigrants;
    } else if (best_EI_fitness < best_DI_fitness) {
      num_rand_immigrants = max(num_rand_immigrants - delta_immigrants, min_immigrants);
      num_elite_immigrants = max(num_elite_immigrants - delta_immigrants, min_immigrants);
      num_dual_immigrants = num_immigrants - num_rand_immigrants - num_elite_immigrants;
    }
  }

  //cout << num_elite_immigrants << " " << num_dual_immigrants 
  //     << " " << num_rand_immigrants << endl;

  pop->mean_fitness = pop->sum_fitness/popsize;

  delete [] vector_order;
}

/*****************************************************************************
 *          		Replacement of part of the population                *
 *****************************************************************************/
void substitution(population *pop) {
  switch (gaMode) {
    case 2:
      imig_random(pop);  // Standard RIs
      break;
    case 3: 
      imig_worst(pop);   // RIs with the replacement of the worst individuals
      break;  
    case 4:
      imig_elitism(pop); // elitism-based immigrants
      break;
    case 5: 
      imig_elitism_random(pop); // elitism-based and random immigrants
      break;
    case 6:
      imig_hybrid(pop);  // hybrid immigrants
      break; 
    default: break;
  }
}
